﻿var esprima = require('esprima'),
    escodegen = require('escodegen'),
    traverse = require('ast-traverse'),
    astUtils = require('esprima-ast-utils'),
    md5 = require('md5'),
    path = require('path'),
    utils = require('../utils'),
    guessMethodName = require('sl-node-cover/lib/util/guess-method-name').guessMethodName,
    consoleFileLogger = require("../sl-logger.js").consoleFileLogger;

var branchNodeTypes = {
    "IfStatement": true,
    "SwitchStatement": true,
    "ConditionalExpression": true,
    "LogicalExpression": true
};

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

function findLogicalExpressionLeaves(node, leaves) {
    if (node.type === "LogicalExpression") {
        findLogicalExpressionLeaves(node.left, leaves);
        findLogicalExpressionLeaves(node.right, leaves);
    } else {
        leaves.push(node);
    }
}

function calculateFunctionNodeSigHash(node) {
    if (!node.params || node.params.length == 0) return "";
    var paramsStr = node.params.map(escodegen.generate).join(',');
    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node, isBranchCoverage) {
    var clonedNode = astUtils.clone(node);
    var branchesData = [];
    //calculate body hash
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression" || node.type == "ArrowFunctionExpression") {
                node.body = {
                    type: "BlockStatement",
                    body: []
                };
                node.params = [];
                node.id = {
                    name: '_'
                };
            }
            if (isBranchCoverage) {
                if (branchNodeTypes[node.type] && node.type !== "LogicalExpression") {
                    addBranchesData(node, branchesData);
                }
                if (node.type == "LogicalExpression" && (parent == undefined || parent.type !== "LogicalExpression")) {
                    addBranchesData(node, branchesData);
                }
            }
        }
    });

    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    if (isBranchCoverage) {
        return {
            hash: hash,
            branchesData: branchesData
        }
    }
    return {
        hash: hash
    }
}

/**
 * Taken from eslint - https://github.com/eslint/eslint/blob/183def6115cad6f17c82ef1c1a245eb22d0bee83/lib/eslint.js#L800
 */
function trimShebang(text) {
    return text.replace(/^#!([^\r\n]+)/, function (match, captured) {
        return "//" + captured;
    });
}

function addBranchesData(node, branchesDataArray) {
    switch (node.type) {
        case "IfStatement":
        case "ConditionalExpression":
            var branchIndex = 0;
            if (node.consequent !== null) {
                branchesDataArray.push(createBranchData(node, node.consequent, node.type, branchIndex++));
            }
            if (node.alternate == null) {
                // Istanbul counts empty alternates and takes the consequent's location
                var clonedNode = astUtils.clone(node);
                clonedNode.alternate = {
                    type: "BlockStatement",
                    body: []
                };
                clonedNode.alternate.loc = node.consequent !== null ? node.consequent.loc : node.loc;
                branchesDataArray.push(createBranchData(clonedNode, clonedNode.alternate, node.type, branchIndex++));
            } else {
                branchesDataArray.push(createBranchData(node, node.alternate, node.type, branchIndex++));
            }
            break;
        case "SwitchStatement":
            var branchIndex = 0;
            node.cases.forEach(function (caseNode) {
                if (caseNode !== null) {
                    branchesDataArray.push(createBranchData(node, caseNode, node.type, branchIndex++));
                }
            });
            break;
        case "LogicalExpression":
            var branchIndex = 0;
            var leaves = [];
            findLogicalExpressionLeaves(node, leaves);
            leaves.forEach(function (leafNode) {
                if (leafNode !== null) {
                    branchesDataArray.push(createBranchData(node, leafNode, node.type, branchIndex++));
                }
            });
            break;
    }
}

function createBranchData(parentNode, node, nodeType, branchIndex) {
    var clonedNode = astUtils.clone(node);
    var position = formatLocToArray(node.loc.start);
    var endPosition = formatLocToArray(node.loc.end);
    //In all cases except "ConditionalExpression" Istanbul reports in runtime on the location
    //of the parent node.
    var parentNodeStart = (nodeType === "ConditionalExpression" || nodeType === "LogicalExpression" || nodeType === "SwitchStatement") ? node.loc.start : parentNode.loc.start;
    var parentPosition = formatLocToArray(parentNodeStart)
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression" || node.type == "ArrowFunctionExpression") {
                node.body = {
                    type: "BlockStatement",
                    body: []
                };
                node.params = [];
                node.id = {
                    name: '_'
                };
            }
            switch (node.type) {
                case "IfStatement":
                case "ConditionalExpression":
                    node.consequent = {
                        type: "BlockStatement",
                        body: []
                    };
                    node.alternate = {
                        type: "BlockStatement",
                        body: []
                    };
                    break;
                case "SwitchStatement":
                    node.cases = [];
                    break;
                case "LogicalExpression":
                    node.operator = "";
                    node.left = {
                        type: "Literal",
                        value: null
                    };
                    node.right = {
                        type: "Literal",
                        value: null
                    };
                    break;
            }
        }
    });
    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    return {
        type: nodeType,
        position: position,
        endPosition: endPosition,
        parentPosition: parentPosition,
        hash: hash,
        index:branchIndex
    };
}

function fixRelPath(p1, p2, logger) {
    var log = logger ? logger.child({
        methodName: "fixRelPath"
    }) : consoleFileLogger("fixRelPath");
    try {
        if (!p1 || p1.length == 0) return p1;
        if (!p2 || p2.length == 0) return p1;

        //Make sure both paths are unix-style

        p1 = utils.adjustPathSlashes(p1);
        p2 = utils.adjustPathSlashes(p2);

        if (p2[0] == '/') return p2; //Absolute path

        var parts1 = p1.split('/');
        parts1.pop(); //remove the filename

        var parts2 = p2.split('/');

        var p3 = parts1.concat(parts2);
        var finalPath = utils.adjustPathSlashes(path.normalize(p3.join('/')));

        return finalPath;
    } catch (err) {
        log.error("Failed to fix real path for source maps. Error: " + err + ". Original File Name: " + p1 + ". Function: " + p2);
        return p2;
    }
}

function excludeOriginalFile(fn) {
    if (fn.indexOf("webpack:/webpack/") === 0) {
        return true; //Embedded resource that has no corresponding source
    }
    return false;
}

function trimPackerPrefix(fn) {
    if (fn.indexOf('webpack:///') === 0) {
        fn = fn.substring(11); //'webpack:///'.length
    }
    return fn;
}

function excludeOriginalFileAfterTrimmingPrefix(fn) {
    if (fn.indexOf("~") >= 0) return true; //Ignore Embedded resource
    if (fn.indexOf(":") >= 0) return true; //Ignore Embedded resource
    if (fn.indexOf("(webpack)") == 0) return true; //Ignore Embedded resource
    return false;
}

function replaceWithOriginalPosition(position, sourceMaps) {
    var originalPosition = sourceMaps.originalPositionFor({
        line: position[0],
        column: position[1]
    });
    if (originalPosition && originalPosition.line !== null && originalPosition.column !== null) {
        return formatLocToArray(originalPosition);
    }
    return position;
}

function generateFileSignature(filename, source, sourceMaps, isBranchCoverage, opts, logger) {
    var result = {
        module: {
            hash: ""
        },
        methods: [],
        branches: []
    };
    var log = logger ? logger.child({
        methodName: "generateFileSignature"
    }) : consoleFileLogger("generateFileSignature");

    filename = utils.adjustPathSlashes(filename); //we want all transients to use unix / instead of windows \

    var ast;

    try {
        source = trimShebang(source);

        //1. Parse the file
        ast = esprima.parse(source, {
            loc: true,
            range: true,
            comment: false, //No need for those,
            sourceType: (opts && opts.es6Modules) ? 'module' : 'script' //Support ES6 non-default 'module' mode
        });

    } catch (parseException) {
        //TODO: Test
        log.error("Failed to parse AST of " + filename + ". " + parseException);
        return {
            error: "Parse error (1)"
        };
    }

    try {

        // print AST node types, pre-order (node first, then its children) 
        traverse(ast, {
            pre: function (node, parent, prop, idx) {
                if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression" || node.type == "ArrowFunctionExpression") {
                    node.isFunctionNode = true;
                    var methodHashObject = calculateFunctionNodeHash(node, isBranchCoverage),
                        hash = methodHashObject.hash,
                        sigHash = calculateFunctionNodeSigHash(node),
                        start = formatLocToArray(node.loc.start),
                        end = formatLocToArray(node.loc.end),
                        nameAndType = guessMethodName(node, parent),
                        name = nameAndType.name, //the actual name, the guessed name, or "(Anonymous)"
                        type = nameAndType.type,
                        isAnonymous = nameAndType.isAnonymous,
                        fn = filename,
                        originalFilename,
                        originalName,
                        mappedPosition,
                        originalStart,
                        originalEnd,
                        originalStartPosition, originalEndPosition;

                    if (sourceMaps) {
                        try {
                            originalStartPosition = sourceMaps.originalPositionFor({
                                line: start[0],
                                column: start[1]
                            });
                            if (!originalStartPosition || !originalStartPosition.source || originalStartPosition.line === null || originalStartPosition.column === null) {
                                return;
                            }
                            originalFilename = fn;
                            fn = originalStartPosition.source;

                            if (excludeOriginalFile(fn)) return;
                            fn = trimPackerPrefix(fn);
                            if (excludeOriginalFileAfterTrimmingPrefix(fn)) return;

                            fn = fixRelPath(originalFilename, fn, log);

                            if (originalStartPosition.name) {
                                originalName = nameAndType.actualName;
                                name = originalStartPosition.name;
                                if (!node.id)
                                    node.id = {
                                        name: originalStartPosition.name
                                    };
                                else if (!node.id.name)
                                    node.id.name = originalStartPosition.name;
                            }

                            originalStart = start;
                            start = [originalStartPosition.line, originalStartPosition.column];

                            originalEndPosition = sourceMaps.originalPositionFor({
                                line: end[0],
                                column: end[1] - 1
                            });
                            originalEnd = end;
                            end = [originalEndPosition.line, originalEndPosition.column];

                        } catch (e) {
                            log.error('Error calculating source maps:' + e);
                        }
                    }

                    // This has to match to the footprints
                    var uniqueName = [name, fn, start.join(',')].join('@');

                    var method = {
                        name: name, //Method that can be presented to the user
                        isAnonymous: isAnonymous,
                        start: start,
                        end: end,
                        hash: hash,
                        sigHash: sigHash,
                        uniqueName: uniqueName,
                        filename: fn,
                        originalFilename: originalFilename,
                        originalStart: originalStart,
                        originalEnd: originalEnd,
                        originalName: originalName,
                        type: type
                    };

                    if (isBranchCoverage) {
                        if (sourceMaps) {
                            methodHashObject.branchesData.forEach(function (branchData) {
                                branchData.position = replaceWithOriginalPosition(branchData.position, sourceMaps);
                                branchData.endPosition = replaceWithOriginalPosition(branchData.endPosition, sourceMaps);
                                branchData.parentPosition = replaceWithOriginalPosition(branchData.parentPosition, sourceMaps);
                                branchData.originalFilename = fn;
                            })
                        }

                        method.branches = methodHashObject.branchesData;      
                    }
                    result.methods.push(method);
                }
                if (node.hasOwnProperty("isFunctionNode") ||
                    (parent !== null && parent.hasOwnProperty("isFunctionNode"))) {
                    // We want to tag nodes that have a parent that is a method
                    node.isFunctionNode = true;
                }
                if (branchNodeTypes[node.type] &&
                    (!node.hasOwnProperty("isFunctionNode") || node.isFunctionNode !== true)) {
                    // Nodes that are of branch types and do not belong to a method parent are global branches
                    //TODO Global branches
                }
            }
        });
    } catch (traverseException1) {
        log.error("Failed to traverse ast. Filename: " + filename + ". Error: " + traverseException1);
        return {
            error: 'Parse error(2)'
        }
    }

    try {
        var bodyHash = calculateFunctionNodeHash(ast).hash;
        result.module.hash = bodyHash;

    } catch (traverseException2) {
        log.error("Failed to calculate module hash. Filename: " + filename + ". Error: traverseException2");
        return {
            error: "Parse error (3)"
        };
    }

    return result;
}

module.exports = generateFileSignature;