﻿var glob = require("glob"),
    Promise = require('promise'),
    path = require('path'),
    fs = require('fs'),
    fileSignature = require('./file-signature.js'),
    methodKeyCreator = require('./method-key.js'),
    branchKeyCreator = require('./branch-key.js'),
    uuid = require('uuid'),
    //statuses = require('../statuses.js'),
    util = require('util'),
    parser = require("gitignore-parser"),
    utils = require('../utils'),
    consoleFileLogger = require("../sl-logger.js").consoleFileLogger,
    ActualFile = require('../file-system/actual-file'),
    InMemoryFile = require('../file-system/in-memory-file');


function generateNewMappingForFile(fileObject, isBranchCoverage, es6Modules, logger) {
    var filename = fileObject.getFilename();
    filename = utils.adjustPathSlashes(filename);
    var fileSig = {
        filename: filename,
        methods: []
    };

    var oldContent = fileObject.getContent();
    var sourceMaps = fileObject.readSourceMaps(logger);
    var sig = fileSignature(filename, oldContent, sourceMaps, isBranchCoverage, {
        es6Modules: es6Modules
    }, logger);

    if (!sig || sig.error) {
        return null;
    }
    fileSig.hash = sig.module.hash;
    fileSig.hasSourceMaps = !!sourceMaps;

    var mapping = {};
    sig.methods.forEach(function (m) {
        var wholeKey = [methodKeyCreator(m.start, m.name, m.isAnonymous), m.hash, new Date().valueOf()].join('|');
        var newId = uuid(wholeKey);
        fileSig.methods.push({
            id: newId,
            isAnonymous: m.isAnonymous,
            hash: m.hash,
            sigHash: m.sigHash,
            type: m.type,
            position: m.start,
            endPosition: m.end,
            name: m.name,
            uniqueName: m.uniqueName,
            filename: m.originalFilename,
            originalFilename: m.filename,
            originalStart: m.originalStart,
            originalEnd: m.originalEnd,
            originalName: m.originalName,
            branches: m.branches
        })
    });
    return fileSig;
}

function findSlIgnore(workspacePath, lookInParentFolders, logger) {
    var searchedPaths = {};
    do {
        var slIgnoreFilePath = path.resolve(workspacePath, '.slignore');
        if (searchedPaths[slIgnoreFilePath]) return null; //already searched at

        logger.debug('Looking for .slignore file at ' + slIgnoreFilePath);
        searchedPaths[slIgnoreFilePath] = true;
        try {
            var f = fs.statSync(slIgnoreFilePath);
            if (f.isFile()) return slIgnoreFilePath;
        } catch (e) {
            workspacePath = workspacePath + '/..';
        }
        if (!lookInParentFolders) break;
    } while (workspacePath.length > 1);
    logger.info('.slignore file not found');
    return null;
}

function getSlIgnore(workspacePath, outputPath, lookInParentFolders, logger) {
    var log = logger ? logger.child({
        methodName: "getSlIgnore"
    }) : consoleFileLogger("getSlIgnore");

    var pattern;
    try {
        var slIgnoreFilePath = findSlIgnore(workspacePath, lookInParentFolders, logger);
        if (!slIgnoreFilePath) throw new Error("SLIGNORE NOT FOUND");
        pattern = fs.readFileSync(slIgnoreFilePath, 'utf8');
        pattern = addExcludedPath(pattern, outputPath);
        try {
            var compiled = parser.compile(pattern); //may throw if the file is invalid/corrupted
            log.info('Using .slignore file ' + slIgnoreFilePath);
            return compiled;
        } catch (e) {
            log.error('Not using corrupted .slignore file ' + slIgnoreFilePath + ". Using default behavior.");
            throw e;
        }
    } catch (e) {
        // default pattern - all js files under test directory and it's sub directories.
        pattern = "";
        pattern = addExcludedPath(pattern, outputPath);
        return parser.compile(pattern); //should not throw again
    }
}

function addExcludedPath(pattern, excludedFolder) {
    if (!excludedFolder) return pattern;
    if (excludedFolder && excludedFolder[excludedFolder.length - 1] != '/') { //Ensure the excludedFolder ends with / to comply with gitignore
        excludedFolder += '/';
    }
    pattern = pattern + "\n" + excludedFolder;
    return pattern;
}

function mapSingleFileInMemory(singleFileParams, es6Modules, logger) {
    var useBranchCoverage = singleFileParams.usebranchcoverage;
    var fileObject = new InMemoryFile(singleFileParams.filename, singleFileParams.originalFilename, singleFileParams.buffer);
    var fileSig = generateNewMappingForFile(fileObject, useBranchCoverage, es6Modules, logger);
    return fileSig;
}

function javascriptHandler(handlerParams, logger) {
    var opts = handlerParams.buildArguments;

    //V1 protocol
    var log = logger ? logger.child({
        methodName: "javascriptHandler"
    }) : consoleFileLogger("javascriptHandler");
    var opts = handlerParams.buildArguments;
    var slIgnore = getSlIgnore(opts.workspacepath, opts.outputpath, !!opts.instrumentForBrowsers, log);
    var finalOutput = handlerParams.finalOutput;

    var p = new Promise(function (resolve, reject) {
        finalOutput.files = finalOutput.files || [];
        var files = [];
        var filesLookup = {};
        var useSlIgnore = true;
        if (handlerParams.buildArguments.specificFiles) {
            files = handlerParams.buildArguments.specificFiles;
            useSlIgnore = false;
        } else {
            handlerParams.buildArguments.includedFiles.forEach(function (inclusionPattern) {
                var matchedFiles = glob.sync(inclusionPattern, {
                    cwd: opts.workspacepath,
                    ignore: handlerParams.buildArguments.excludedFiles
                });
                matchedFiles.forEach(function (matchedFile) {
                    if (!filesLookup[matchedFile]) {
                        filesLookup[matchedFile] = true;
                        files.push(matchedFile);
                    }
                });
            });

        }
        handlerParams.processedFiles = [];
        var erroredFiles = [];

        files.forEach(function (filename, fileIdx) {
            var fullFilename = path.resolve(opts.workspacepath, filename);

            // If the file is filtered by the .slignore file:
            if (useSlIgnore && slIgnore.denies(filename)) {
                log.debug('Excluded [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);
                return; // Ignored by .slignore, skip it.
            }
            handlerParams.processedFiles.push(filename);

            log.info('Processing [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);

            var useBranchCoverage = handlerParams.cfg.useBranchCoverage || opts.usebranchcoverage;

            var fileObject = new ActualFile(filename, fullFilename);
            var fileSig = generateNewMappingForFile(fileObject, useBranchCoverage, opts.es6Modules || opts["es-modules"], logger);

            //If source maps are used, the methods in fileSig may originate from excluded files

            if (!useSlIgnore) {
                var includedMethods = [];
                fileSig.methods.forEach(function (m) {
                    if (m.filename) {
                        if (!slIgnore.denies(m.filename)) {
                            includedMethods.push(m);
                            delete m.filename;
                        } else {
                            //console.log('Rejecting method from file: '+m.filename);
                        }
                    } else {
                        includedMethods.push(m);
                    }
                });

                fileSig.methods = includedMethods;

                if (fileSig.branches){
                    var includedBranches = [];
                    fileSig.branches.forEach(function(branch){
                        if (branch.filename) {
                            if (!slIgnore.denies(branch.filename)){
                                includedBranches.push(branch);
                                delete branch.filename;
                            }
                            else {
                                //console.log('Rejecting branch from file: '+branch.filename);
                            }
                        } else{
                            includedBranches.push(branch);
                        }
                    });

                    fileSig.branches = includedBranches;
                }

            } else {
                if (fileSig && fileSig.methods) {
                    fileSig.methods.forEach(function (m) {
                        delete m.filename;
                    });
                }

                if (fileSig && fileSig.branches) {
                    fileSig.branches.forEach(function (b) {
                        delete b.filename;
                    });
                }


                if (fileSig) {
                    finalOutput.files.push(fileSig);
                }
                else {
                    erroredFiles.push(filename);
                }
            }
        });

        if (erroredFiles.length) {
            log.warn(
                "The following files failed to be parsed and will not be sent to the server:\n" +
                erroredFiles.join('\n') +
                "\n\t(Hint: if the file use the es6 module syntax, pass the \"--es-modules\" switch)");
            return resolve(false); //Don't fail the build scanning process if some files failed to be parsed
        } else {
            return resolve(true);
        }
    });
    return p;
}

function javascriptHandlerV2(handlerParams, logger) {
    return javascriptHandler(handlerParams, logger).then(function () {
        var p = new Promise(function (resolve, reject) {
            var log = logger ? logger.child({
                methodName: "javascriptHandlerV2"
            }) : consoleFileLogger("javascriptHandlerV2"),
                opts = handlerParams.buildArguments,
                finalOutput = handlerParams.finalOutput,

                unsortedFiles = finalOutput.files,
                filesLookup = {},
                filesArray = [];
            finalOutput.files = filesArray;
            finalOutput.version = "2.0";

            function getCreateFile(filename) {
                if (!filesLookup[filename]) {
                    var newFile = {
                        logicalPath: filename,
                        physicalPath: filename,
                        methods: [],
                        lines: [],
                        branches: []
                    };
                    filesLookup[filename] = newFile;
                    filesArray.push(newFile);
                }
                return filesLookup[filename];
            }

            unsortedFiles.forEach(function (file) {
                if (!file) return;
                if (file.methods) {
                    file.methods.forEach(function (method) {
                        if (!method) return;
                        var trueSrcFilename = method.originalFilename || file.filename;
                        if (file.hasSourceMaps && trueSrcFilename == file.filename) { // Skip method that is not mapped to any "true" source file. It's compiler generated
                            return;
                        }
                        var methodFile = getCreateFile(trueSrcFilename);

                        var convertedMethod = {
                            uniqueId: [trueSrcFilename, method.position.join(',')].join('@'),
                            displayName: method.name,
                            position: method.position,
                            endPosition: method.endPosition,
                            hash: method.hash,
                            sigHash: method.sigHash,
                            meta: {} //Filled-in below
                        };
                        if (method.isAnonymous) {
                            convertedMethod.meta.anonymous = true;
                        }
                        if (method.type) {
                            convertedMethod.meta.type = method.type;
                        }
                        methodFile.methods.push(convertedMethod);
                        
                        if (!file.branches)
                            file.branches = [];

                        if (method.branches) {
                            method.branches.forEach(function (branch) {
                                if (!branch) return;
                                //TODO: Hack! Currently the branches which we collect are inside method.
                                //When we'll support global branches, we should get the "original file name"
                                //in the same time we create the branch data object.
                                branch.originalFilename = method.originalFilename; 
                                file.branches.push(branch);
                            });
                        }
                    })
                }

                if (file.branches){
                    file.branches.forEach(function (branch) {
                        if (!branch) return;
                        var trueSrcFilename = branch.originalFilename || file.filename;
                        if (file.hasSourceMaps && trueSrcFilename == file.filename) { // Skip method that is not mapped to any "true" source file. It's compiler generated
                            return;
                        }
                        var branchFile = getCreateFile(trueSrcFilename);

                        var uniqueId = branchKeyCreator(branch.originalFilename, branch.parentPosition, branch.index);
                        var b = {
                            uniqueId:uniqueId,
                            filename: branch.filename,
                            type: branch.type,
                            position: branch.position,
                            endPosition: branch.endPosition,
                            hash: branch.hash,
                            index: branch.index
                        }
                        branchFile.branches.push(b);
                    });
                }
            });

            return resolve(true);
        });
        return p;
    });
}

module.exports.javascriptHandler = javascriptHandler;
module.exports.javascriptHandlerV2 = javascriptHandlerV2;
module.exports.mapSingleFileInMemory = mapSingleFileInMemory;